/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "network/packet/packetqueue.h"
#include "networkclient.h"

/**
 * Type of connection used for a client.
 */
typedef enum {
  CLIENT_TYPE_SINGLE_PLAYER = 0,
  CLIENT_TYPE_NETWORKED = 1,
} clienttype_t;

/**
 * The state of the client and what it is doing.
 */
typedef enum {
  CLIENT_STATE_DISCONNECTED = 0,
  CLIENT_STATE_CONNECTING,
  CLIENT_STATE_CONNECTED,
  CLIENT_STATE_DISCONNECTING,
} clientstate_t;

/**
 * Information used to connect to a server.
 */
typedef struct clientconnect_s {
  clienttype_t type;
} clientconnect_t;

/**
 * Error codes for the client.
 */
typedef enum {
  CLIENT_ERROR_NO_ERROR = 0,
  CLIENT_ERROR_CONNECT_THREAD_SPAWN_FAILED,
} clienterror_t;

/**
 * The clients state and information.
 */
typedef struct client_s {
  clientstate_t state;
  clienttype_t type;
  packetqueue_t queue;
  clienterror_t error;

  pthread_t connectThread;

  union {
    networkclient_t network;
  };
} client_t;

extern client_t CLIENT;

/**
 * Initializes the client object, does not perform any connection.
 */
void clientInit();

/**
 * Initiates a client connection, this function will not block and will return
 * immediately even if a connection issue occurs. It is up to the caller to
 * handle the connection state and any errors that occur.
 * 
 * Client must be already disconnected before calling this function.
 * 
 * @param connect The connection information to use.
 */
void clientConnect(const clientconnect_t connect);

/**
 * Requests the client to disconnect from the server. This will not block and
 * will return immediately. The client will handle the disconnection in the
 * background.
 */
void clientDisconnect();

/**
 * Internal method that writes a packet to the client immediately.
 */
clienterror_t clientWritePacket(const packet_t *packet);

/**
 * Internal method that reads a packet from the client immediately.
 * 
 * @param packet The packet to read into.
 * @return Error code indicating success or failure.
 */
clienterror_t clientReadPacket(const packet_t *packet);

/**
 * Main update loop for the client, this will handle all incoming packets and
 * update the client state.
 */
void clientUpdate();

/**
 * Disposes of the client, cleans up any resources and will force disconnect if
 * not already disconnected.
 */
void clientDispose();

/**
 * Thread function responsible for initiating and handshaking the connection.
 * 
 * @param arg The client connection information.
 * @return Returns NULL, unless an error occurs.
 */
void * clientThreadConnect(void *arg);